# __init__.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a NoSQL database of chess games using the vedis package.

This interface can be used if the vedis package is installed, from PyPI for
example.
"""
