# __init__.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a Berkeley DB database of chess games using tkinter Tcl interface.

This interface can be used if installed Berkeley DB is built with Tcl support.
"""
